/*
 * ROX-Filer, filer for the ROX desktop project
 * By Thomas Leonard, <tal197@users.sourceforge.net>.
 */

#ifndef _INFOBOX_H
#define _INFOBOX_H

void infobox_show_list(GList *paths);
void infobox_new(const gchar *pathname);

#endif /* _INFOBOX_H */
