/*
 * ROX-Filer, filer for the ROX desktop project
 * By Thomas Leonard, <tal197@users.sourceforge.net>.
 */

#ifndef _TASKLIST_H
#define _TASKLIST_H

void tasklist_set_active(gboolean active);
void tasklist_style_changed(void);

#endif /* _TASKLIST_H */
