/*
 * ROX-Filer, filer for the ROX desktop project
 * By Thomas Leonard, <tal197@users.sourceforge.net>.
 */

#ifndef _CELL_ICON_H
#define _CELL_ICON_H

GtkCellRenderer *cell_icon_new(ViewDetails *view_details);

#endif /* _CELL_ICON_H */
